char *Version = "2.99.28 #1: Tue Jul 16 03:28:18 EET DST 1996";
char *VersionNumb = "2.99.28 #1";
char *CC_user = "mea@mea2.pp.utu.fi";
char *CC_pwd = "/usr/home/mea/zmailer-2.99.28/transports/smtp";
